# treeNobj.tcl --
#	This file implements a splay tree data structure for Tcl
#	programs as an object.
#
# Copyright (c) 1998 Clif Flynt 
#
# This program is free for use in non-commercial applications, provided 
# that the copyright information is retained.  It may be licensed from
# Flynt Consulting Services for use in commercial applications.    
#
#  This code is described in "Tcl/Tk for Real Programmers", 
#  published by Academic Press Professional
#  ISBN: 0122612051 
#
# If -testObj is only command line arg, then this is being run standalone
#    for a test.
#

if {([info exists argv]) && ($argv == "-testObj" )} {
    lappend auto_path "."
}

set treeObject_version [lindex {$RCSfile: treeNobj.tcl,v $ $Revision: 1.20 $} 4]

package provide treeObject 2.2

########################################################################
# proc newTreeObject {} --
#  Creates a new Tree object, and returns an object for accessing the
#    base node.
# Arguments:
#   NONE

proc newTreeObject {} {
    set treeId [treeObject::tree::treeCreate]

    proc $treeId {args} "
	return \[treeObject::TreeObjProc $treeId \$args]
    "

    return $treeId
}

namespace eval treeObject {
package require tree 2.0
if {[info exists auto_index(::tree::treeCreate)]} {
    eval $auto_index(::tree::treeCreate)
    namespace import tree::*
}

########################################################################
# proc TreeObjProc {node methodArgs} --
#    Primary method for tree objects.  
#    Parses the subcommand and calls the appropriate tree.tcl proc.
#
# Arguments:
#    node:	  The internal name for this node.
#    methodArgs:  A list, the first member is the subcommand name,
#		  subsequent (optional) members are arguments to be
#		  passed to the tree.tcl command.
#
# Results:
#    Calls other procedures which may produce output, create new objects,
#	delete objects, etc.

proc TreeObjProc {node methodArgs} {

    set subCommand [lindex $methodArgs 0]
    set cmdArgs [lrange $methodArgs 1 end]

    switch $subCommand {
    {set} {
	    # Set the Key/Value pair associated with this node.
	    # could be set -node key value
	    #          set -tree key value
	    #          set key value   -- Default to node.
	    if {[llength $cmdArgs] == 3} {
		set option [lindex $cmdArgs 0]
		set key [lindex $cmdArgs 1]
		set val [lindex $cmdArgs 2]
	    }\
	    elseif {[llength $cmdArgs] == 2} {
		set option "-node"
		set key [lindex $cmdArgs 0]
		set val [lindex $cmdArgs 1]
	    } else {
		error "$node set ?-node? ?-tree? key value"
	    }
	    if {[string match $option "-node"]} {
		return [treeSetNodeValue $node $key $val]
	    }\
	    elseif {[string match $option "-tree"]} {
		return [treeSetTreeValue $node $key $val]
	    } else {
		error "$node set ?-node? ?-tree? key value"
	    }
	}
    {append} {
	    # Append(or assign)  data to the Key/Value pair associated 
	    #  with this node or tree.
	    # could be append -node key value
	    #          append -tree key value
	    #          set key value   -- Default to node.
	    if {[llength $cmdArgs] == 3} {
		set option [lindex $cmdArgs 0]
		set key [lindex $cmdArgs 1]
		set val [lindex $cmdArgs 2]
	    }\
	    elseif {[llength $cmdArgs] == 2} {
		set option "-node"
		set key [lindex $cmdArgs 0]
		set val [lindex $cmdArgs 1]
	    } else {
		error "$node set ?-node? ?-tree? key value"
	    }
	    if {[string match $option "-node"]} {
		return [treeSetNodeValue $node $key $val -append]
	    }\
	    elseif {[string match $option "-tree"]} {
		return [treeSetTreeValue $node $key $val -append]
	    } else {
		error "$node set ?-node? ?-tree? key value"
	    }
	}
    {lappend} {
	    # Append(or assign)  data to the Key/Value pair associated 
	    #  with this node or tree.
	    #          lappend -tree key value
	    #          lappend key value   -- Default to node.
	    if {[llength $cmdArgs] == 3} {
		set option [lindex $cmdArgs 0]
		set key [lindex $cmdArgs 1]
		set val [lindex $cmdArgs 2]
	    }\
	    elseif {[llength $cmdArgs] == 2} {
		set option "-node"
		set key [lindex $cmdArgs 0]
		set val [lindex $cmdArgs 1]
	    } else {
		error "$node set ?-node? ?-tree? key value"
	    }
	    if {[string match $option "-node"]} {
		return [treeSetNodeValue $node $key $val -lappend]
	    }\
	    elseif {[string match $option "-tree"]} {
		return [treeSetTreeValue $node $key $val -lappend]
	    } else {
		error "$node set ?-node? ?-tree? key value"
	    }
	}
    {get} {
	    # return the Key/Value pair associated with this node.
	    #  May be:
	    #  get -node key
	    #  get -tree key
	    #  get -best key  -- return node, or tree if no node value
	    #  get key	     -- treat as best.

	    if {[llength $cmdArgs] == 2} {
		set option [lindex $cmdArgs 0]
		set key [lindex $cmdArgs 1]
	    }\
	    elseif {[llength $cmdArgs] == 1} {
		set option "-best"
		set key [lindex $cmdArgs 0]
	    } else {
		error "$node get ?-node? ?-tree? key"
	    }
	    switch -exact -- $option {
	    -node {
		    return [treeGetNodeValue $node $key]
		}
	    -tree {
		    return [treeGetTreeValue $node $key]
		}
	    -best {
		    set fail [catch {treeGetNodeValue $node $key} returnVal]
		    if {!$fail} {
			return $returnVal
		    }
		    return [treeGetTreeValue $node $key]
		}
	    default {
		    error "$node get ?-node? ?-tree? key"
		}
	    }
	}
    {unset} {
	    # return the Key/Value pair associated with this node.
	    #  May be:
	    #  unset -node key
	    #  unset -tree key
	    #  unset key	     -- treat as node.

	    if {[llength $cmdArgs] == 2} {
		set option [lindex $cmdArgs 0]
		set key [lindex $cmdArgs 1]
	    }\
	    elseif {[llength $cmdArgs] == 1} {
		set option "-node"
		set key [lindex $cmdArgs 0]
	    } else {
		error "$node unset ?-node? ?-tree? key"
	    }
	    switch -exact -- $option {
	    -node {
		    return [treeUnsetNodeValue $node $key]
		}
	    -tree {
		    return [treeUnsetTreeValue $node $key]
		}
	    default {
		    error "$node unset ?-node? ?-tree? key"
		}
	    }
	}

    {add} {
	    # Add a child node to this node.
	    set newNode [treeCreateChild $node]

	    set str "proc $newNode {args} {
		return \[[namespace current]::TreeObjProc $newNode \$args]
	    }"

	    namespace eval "::" $str
	    return $newNode
	}
    {delete} {
	    # remove this node 
	    # and remove the proc from the proc hash table
	    # If cmdArgs == "-r", delete recursively.
	    if {[string match $cmdArgs "-r"]} {
		foreach child [$node childList] {
		    $child delete -r
		    catch {rename $child ""}
		}
	    }
	    # delete this node.
	    treeDeleteNode $node
	    rename $node ""
	}
    {destroyTree} {
	    # destroy this tree completely
	    # Find the base of the tree, and delete all nodes
	    #  and their associated procs
	    set base [$node base]
	    $base delete -r

	    # then destroy the base.
	    treeDestroyTree $node
	}
    {childList} {
	    # return a list of children
	    return [treeGetChildList $node]
	}
    {search} {
	    # return a list of keys defined in this tree
	    #   that match a pattern
	    return [treeSearchKey $node $cmdArgs]
	}
    {siblingList} {
	    # return a list of children
	    return [treeGetSiblingList $node]
	}
    {parent} {
	    # returns the parent of a node
	    return [treeGetParent $node]
	}
    {valueList} {
            # returns a list of ids and values
            #  May be:
            #  valuelist -node
            #  valuelist -tree
            #  valuelist      ;# treat as node
            if {[llength $cmdArgs] == 1} {
                set option [lindex $cmdArgs 0]
            }\
            elseif {[llength $cmdArgs] == 0} {
                set option -node
            } else {
                error "$node valuelist ?-node? ?-tree?"
            }
            switch -exact -- $option {
            -node {
                    return [treeGetNodeValueList $node]
            }
            -tree {
                    return [treeGetTreeValueList $node]
            }
            default {
                    error "$node valuelist ?-node? ?-tree?"
                }
            }
        }
    {keyList} {
            # returns a list of ids and values
            #  May be:
            #  keyList -node
            #  keyList -tree
            #  keyList      ;# treat as node
            if {[llength $cmdArgs] == 1} {  
                set option [lindex $cmdArgs 0]
            }\
            elseif {[llength $cmdArgs] == 0} {
                set option -node
            } else {
                error "$node keyList ?-node? ?-tree?"
            }
            switch -exact -- $option {
            -node {
                    return [treeGetNodeKeyList $node]
            }
            -tree {
                    return [treeGetTreeKeyList $node]
            }
            default {
                    error "$node keyList ?-node? ?-tree? "
                }
            }
        }
    {base} {
	    # returns the ultimate ancestor of a node.
	    return [treeGetRoot $node]
	}
    {dump} {
	    # dump the contents of this node
	    return [treeDump $node $cmdArgs]
	}
    {getNodes} {
	    # dump the contents of this node
	    return [treeGetAllNodes $node ]
	}
    default {
	    # Should not be here, this is an error
            set methods {set, get, add, delete  destroyTree childList, \
                 parent, keyList, valueList, base, dump, siblingList, search}
	    error "Bad method: $subCommand\n use $methods"
	}
    }
    return ""
}


if {([info exists argv]) && ($argv == "-testObj")} {
proc TestTreeObj {} {
    
    # Test creating trees

    if {![string match [set mytree [newTreeObject]] "tree0.1"]} {
	error "Failed to get expected return from newTreeObject:  $mytree !=\
	  tree0.1"
    }
    
    # Test various methods for setting and getting id/value pairs.

    if {![string match [set tmp [$mytree set -node id1 value1]] "value1"]} {
	error "Failed to get expected return from set -node:  $tmp != value1"
    }

    if {![string match [set tmp [$mytree set -node id2 value2]] "value2"]} {
	error "Failed to get expected return from set -node:  $tmp != value2"
    }

    if {![string match [set tmp [$mytree get -node id1]] "value1"]} {
	error "Failed to get expected return from get -node:  $tmp != value1"
    }

    if {![string match [set tmp [$mytree get -node id2]] "value2"]} {
	error "Failed to get expected return from get -node:  $tmp != value2"
    }
    
    if {![string match [set tmp [$mytree get -best id2]] "value2"]} {
	error "Failed to get expected return from get -best:  $tmp != value2"
    }

    if {![string match [set tmp [$mytree get id2]] "value2"]} {
	error "Failed to get expected return from get:  $tmp != value2"
    }

    set fail [catch {$mytree get -tree id2} tmp]

    if {!$fail || [string match $tmp "value2"]} {
	error "Got un-expected return from get -tree:  $tmp != value2"
    }
    
    # Add some children

    if {![string match [set child [$mytree add]] "tree0.2"]} {
	error "Failed to get expected return from add:  $child != tree0.2"
    }

    if {![string match [set child1 [$mytree add]] "tree0.3"]} {
	error "Failed to get expected return from add:  $child1 != tree0.3"
    }

    if {![string match [set child1_1 [$child1 add]] "tree0.4"]} {
	error "Failed to get expected return from add:  $child1_1 !=\
	  tree0.4"
    }

    if {![string match [set child1_2 [$child1 add]] "tree0.5"]} {
	error "Failed to get expected return from add:  $child1_2 !=\
	  tree0.5"
    }
    
    # Test the key search
    if {![string match [set tmp [$child1_2 search "id2"]] "$mytree"]} {
	error "Failed to get expected return from search:  $tmp !=\
	  $mytree"
    }

    # Test more set/get with multiple word values and keys.

    if {![string match [set t [$child1 set -node id6 "this's a multiword\
      string"]] "this's a multiword string"]} {
	error "Failed to get expected return from set -node:  $t != this's a\
	  multiword string"
    }

    if {![string match [set t [$child1 get -node id6]] "this's a multiword\
      string"]} {
	error "Failed to get expected return from get -node:  $t != this's a\
	  multiword string"
    }

    if {![string match [set t [$child1 set -node "multi word id"\
      "multiword data"]] "multiword data"]} {
	error "Failed to get expected return from set -node:  $t != multiword\
	  data"
    }

    if {![string match [set t [$child1 get -node "multi word id"]]\
      "multiword data"]} {
	error "Failed to get expected return from get -node:  $t != multiword\
	  data"
    }

    if {![string match [set t [$child1 set -tree "multi word id"\
      "TREE multiword data"]] "TREE multiword data"]} {
	error "Failed to get expected return from set -node:  $t != TREE\
	  multiword data"
    }

    if {![string match [set t [$child1 set -tree "TreeID"\
      "TREE ID data"]] "TREE ID data"]} {
	error "Failed to get expected return from set -node:  $t != TREE\
	  ID data"
    }

    if {![string match [set t [$child1 get -node "multi word id"]]\
      "multiword data"]} {
	error "Failed to get expected return from get -node:  $t != multiword\
	  data"
    }

    if {![string match [set t [$child1 get -best "multi word id"]]\
      "multiword data"]} {
	error "Failed to get expected return from get -best:  $t != multiword\
	  data"
    }

    if {![string match [set t [$child1 get "multi word id"]]\
      "multiword data"]} {
	error "Failed to get expected return from get:  $t != multiword data"
    }

    if {![string match [set t [$child1 get -tree "multi word id"]]\
      "TREE multiword data"]} {
	error "Failed to get expected return from get:  $t != TREE multiword\
	  data"
    }

    if {![string match [set t [$child1_1 get -best "multi word id"]]\
      "TREE multiword data"]} {
	error "Failed to get expected return from get: -best  $t != TREE\
	  multiword data"
    }

    if {![string match [set t [$child1_1 get "multi word id"]] "TREE multiword\
      data"]} {
	error "Failed to get expected return from get:  $t != TREE multiword\
	  data"
    }


    # Test the tree dump 

    set expectDump {{ NODE: tree0.1 - primary: tree0 - idNum: 1}\
      {  Parent: None} {  Values:} {    tree0.1.node.id2 value2}\
      {    tree0.1.node.id1 value1} {  Children: tree0.2 tree0.3}\
      {     NODE: tree0.2 - primary: tree0 - idNum: 2} {      Parent:\
      tree0.1} {      treeGetNodeValueList: Node tree0.2 has no\
      values} {      Children: None} {     NODE: tree0.3 - primary: tree0\
      - idNum: 3} {      Parent: tree0.1} {      Values:}\
      {        tree0.3.node.id6 this's a multiword string}\
      {        tree0.3.node.multi word id multiword data} {      Children:\
      tree0.4 tree0.5} {         NODE: tree0.4 - primary: tree0\
      - idNum: 4} {          Parent: tree0.3}\
      {          treeGetNodeValueList: Node tree0.4 has no values}\
      {          Children: None} {         NODE: tree0.5 - primary: tree0\
      - idNum: 5} {          Parent: tree0.3}\
      {          treeGetNodeValueList: Node tree0.5 has no values}\
      {          Children: None}}
    set dump [$mytree dump -r]

    if {$dump != $expectDump} {

	for {set i 0} {$i < [llength $dump]} {incr i} {
	    puts "[string match [lindex $dump $i] [lindex $expectDump $i]]: \n\
	      [lindex $dump $i] \n   [lindex $expectDump $i] "
	}
	error "Failed to get expected return from dump:  \n..$dump..\n !=\
	  \n..$expectDump..\n"

    }
    
    # Test getNodes
    
    set expectList {tree0.1 tree0.2 tree0.3 tree0.4 tree0.5}
    
    set got [$mytree getNodes]

    if {![string match $expectList $got]} {
	error "Failed to get expected return from \
	getNodes \n$got \n!= \n$expectList"
    }

    # Test childList

    if {![string match [set t [$child1 childList]]\
      "tree0.4 tree0.5"]} {
	error "Failed to get expected return from childList:  $t !=\
	  tree0.4 tree0.5"
    }
    
    # Test parent

    if {![string match [set t [$child1_1 parent]] $child1]} {
	error "Failed to get expected return from parent:  $t != $child1"
    }
    
    # test valueList

    set expectValueList {{id6 {this's a multiword string}}\
      {{multi word id} {multiword data}}}

    if {![string match [set t [$child1 valueList]] $expectValueList]} {
	error "Failed to get expected return from valueList:  \n..$t.. \n!=\
	  \n..$expectValueList"
    }

    if {![string match [set t [$child1 valueList -node]] $expectValueList]} {
	error "Failed to get expected return from valueList -node:  \n..$t.. \n!=\
	  \n..$expectValueList"
    }

    set expectValueList \
        {{{multi word id} {TREE multiword data}} {TreeID {TREE ID data}}}

    if {![string match [set t [$child1 valueList -tree]] $expectValueList]} {
	error "Failed to get expected return from valueList -tree:  \n..$t.. \n!=\
	  \n..$expectValueList"
    }

    set expectKeyList {id6 {multi word id}}

    if {![string match [set t [$child1 keyList]] $expectKeyList]} {
	error "Failed to get expected return from keyList:  \n..$t.. \n!=\
	  \n..$expectKeyList"
    }

    if {![string match [set t [$child1 keyList -node]] $expectKeyList]} {
	error "Failed to get expected return from keyList -node:  \n..$t.. \n!=\
	  \n..$expectKeyList"
    }


    set expectKeyList {{multi word id} TreeID}

    if {![string match [set t [$child1 keyList -tree]] $expectKeyList]} {
	error "Failed to get expected return from keyList -tree:  \n..$t.. \n!=\
	  \n..$expectKeyList"
    }
    
    # test unset.

    set expectKeyList {{multi word id}}
    $child1 unset -node id6

    if {![string match [set t [$child1 keyList]] $expectKeyList]} {
	error "Failed to get expected return from keyList:  \n..$t.. \n!=\
	  \n..$expectKeyList"
    }


    
    set expectKeyList {TreeID}
    $child1 unset -tree {multi word id}

    if {![string match [set t [$child1 keyList -tree]] $expectKeyList]} {
	error "Failed to get expected return from keyList -tree:  \n..$t.. \n!=\
	  \n..$expectKeyList"
    }
    
    # test lappend and append 
    
    set expected {{lapp1 {{This is a test} {second list}}}\
    	{app1 {This is a testsecond string}}}
    
    $child append app1 "This is a test"
    $child lappend lapp1 "This is a test"

    $child append app1 "second string"
    $child lappend lapp1 "second list"

    if {![string match [set t [$child valueList]] $expected]} {
	error "Failed to get expected return from append/lappend test:  \n..$t.. \n!=\
	  \n..$expected.."
    }

    if {![string match [set t [$child valueList -node]] $expected]} {
	error "Failed to get expected return from -node append/lappend test:  \n..$t.. \n!=\
	  \n..$expected.."
    }
    
    $child unset app1
    $child unset lapp1

    set expected {{lapp1 {{This is a test} {second list}}}\
    	{app1 {This is a testsecond string}}}
    
    $child append -tree tree1 "This is a test" 
    $child lappend -tree tree2 "This is a test" 

    $child append -tree tree1 "second string"
    $child lappend -tree tree2 "second list"

    set expected ""

    if {![string match [set t [$child valueList]] $expected]} {
	error "Failed to get expected return from append/lappend test:  \
	\n..$t.. \n!= \n..$expected.."
    }

    set expected {{tree1 {This is a testsecond string}}\
        {tree2 {{This is a test} {second list}}} {TreeID {TREE ID data}}}

    if {![string match [set t [$child valueList -tree]] $expected]} {
	error "Failed to get expected return from -tree append/lappend test:\
	   \n..$t.. \n!= \n..$expected.."
    }
    
    # test base

    if {![string match [set t [$child1 base]] $mytree]} {
	error "Failed to get expected return from base:  $t != $mytree"
    }

    # test methods of deleting nodes and trees

    $child1 delete -r

    set fail [catch {puts "[$child1_2 set -node id4]"}]
    if {!$fail} {
	error "Successful set in a deleted node"
    }

    set fail [catch {puts "[$child1 set -node id4]"}]
    if {!$fail} {
	error "Successful set in a child of a deleted node"
    }

    $mytree destroyTree
    set fail [catch {puts "[$child set -tree id3]"}]
    if {!$fail} {
	error "Set an entry in a deleted node!"
    }
    
    # Confirm a new tree can be constructed.

    set fail [catch {newTreeObject} mytree2]
    if {$fail} {
	error "failed to recreate a destroyed tree!"
    }

    if {![string match [set strange [newTreeObject]] "tree2.1"]} {
	error "Failed to get expected return from newTreeObject:  $t !=\
	  $tree2.1"
    }
    

    puts "tree Object tests OK - No Errors"

}
TestTreeObj
}
}


