# treeObj.tcl --
#	This file implements a splay tree data structure for Tcl
#	programs as an object.
#
# Copyright (c) 1998 Clif Flynt 
#
# This program is free for use in non-commercial applications, provided 
# that the copyright information is retained.  It may be licensed from
# Flynt Consulting Services for use in commercial applications.    
#
#  This code is described in "Tcl/Tk for Real Programmers", 
#  published by Academic Press Professional
#  ISBN: 0122612051 
#
# If -test is only command line arg, then this is being run standalone
#    for a test.
#


set treeObj_version [lindex {$RCSfile: treeObj.tcl,v $ $Revision: 1.23 $} 4]

source tree.tcl

########################################################################
# proc newTreeObject {} --
#  Creates a new Tree object, and returns an object for accessing the
#    base node.
# Arguments:
#   NONE

proc newTreeObject {} {
    set treeId [treeCreate]
    proc $treeId {args} "
	return \[TreeObjProc $treeId \$args]
    "

    return $treeId
}

########################################################################
# proc TreeObjProc {node methodArgs} --
#    Primary method for tree objects.  
#    Parses the subcommand and calls the appropriate tree.tcl proc.
#
# Arguments:
#    node:	  The internal name for this node.
#    methodArgs:  A list, the first member is the subcommand name,
#		  subsequent (optional) members are arguments to be
#		  passed to the tree.tcl command.
#
# Results:
#    Calls other procedures which may produce output, create new objects,
#	delete objects, etc.

proc TreeObjProc {node methodArgs} {

    set subCommand [lindex $methodArgs 0]
    set cmdArgs [lrange $methodArgs 1 end]

    switch $subCommand {
    {set} {
	    # Set the Key/Value pair associated with this node.
	    # could be set -node key value
	    #          set -tree key value
	    #          set key value   -- Default to node.
	    if {[llength $cmdArgs] == 3} {
		set option [lindex $cmdArgs 0]
		set key [lindex $cmdArgs 1]
		set val [lindex $cmdArgs 2]
	    }\
	    elseif {[llength $cmdArgs] == 2} {
		set option "-node"
		set key [lindex $cmdArgs 0]
		set val [lindex $cmdArgs 1]
	    } else {
		error "$node set ?-node? ?-tree? key value"
	    }
	    if {[string match $option "-node"]} {
		return [treeSetNodeValue $node $key $val]
	    }\
	    elseif {[string match $option "-tree"]} {
		return [treeSetTreeValue $node $key $val]
	    } else {
		error "$node set ?-node? ?-tree? key value"
	    }
	}
    {get} {
	    # return the Key/Value pair associated with this node.
	    #  May be:
	    #  get -node key
	    #  get -tree key
	    #  get -best key  -- return node, or tree if no node value
	    #  get key	     -- treat as best.

	    if {[llength $cmdArgs] == 2} {
		set option [lindex $cmdArgs 0]
		set key [lindex $cmdArgs 1]
		set val [lindex $cmdArgs 2]
	    }\
	    elseif {[llength $cmdArgs] == 1} {
		set option "-best"
		set key [lindex $cmdArgs 0]
		set val [lindex $cmdArgs 1]
	    } else {
		error "$node set ?-node? ?-tree? key value"
	    }
	    switch -exact -- $option {
	    -node {
		    return [treeGetNodeValue $node $key]
		}
	    -tree {
		    return [treeGetTreeValue $node $key]
		}
	    -best {
		    set fail [catch {treeGetNodeValue $node $key} returnVal]
		    if {!$fail} {
			return $returnVal
		    }
		    return [treeGetTreeValue $node $key]
		}
	    default {
		    error "$node set ?-node? ?-tree? key value"
		}
	    }
	}
    {add} {
	    # Add a child node to this node.
	    set newNode [treeCreateChild $node]
	    proc $newNode {args} "
		return \[TreeObjProc $newNode \$args]
	    "
	    return $newNode
	}
    {delete} {
	    # remove this node 
	    # and remove the proc from the proc hash table
	    # If cmdArgs == "-r", delete recursively.
	    if {[string match $cmdArgs "-r"]} {
		foreach child [$node childList] {
		    $child delete -r
		    catch {rename $child ""}
		}
	    }
	    # delete this node.
	    treeDeleteNode $node
	    rename $node ""
	}
    {destroyTree} {
	    # destroy this tree completely
	    # Find the base of the tree, and delete all nodes
	    #  and their associated procs
	    set base [$node base]
	    $base delete -r

	    # then destroy the base.
	    treeDestroyTree $node
	}
    {childList} {
	    # return a list of children
	    return [treeGetChildList $node]
	}
    {search} {
	    # return a list of keys defined in this tree
	    #   that match a pattern
	    return [treeSearchKey $node $cmdArgs]
	}
    {siblingList} {
	    # return a list of children
	    return [treeGetSiblingList $node]
	}
    {parent} {
	    # returns the parent of a node
	    return [treeGetParent $node]
	}
    {valueList} {
	    # returns a list of ids and values
	    return [treeGetNodeValueList $node]
	}
    {base} {
	    # returns the ultimate ancestor of a node.
	    return [treeGetRoot $node]
	}
    {dump} {
	    # dump the contents of this node
	    return [treeDump $node $cmdArgs]
	}
    default {
	    # Should not be here, this is an error
            set methods {set, get, add, delete  destroyTree childList, \
                 parent, valueList, base, dump, siblingList search}

	    error "Bad method: $subCommand\n use $methods"
	}
    }
    return ""
}
proc TestTreeObj {} {
    
    # Test creating trees

    if {![string match [set mytree [newTreeObject]] "tree0.node.1"]} {
	error "Failed to get expected return from newTreeObject:  $mytree !=\
	  tree0.node.1"
    }
    
    # Test various methods for setting and getting Key/Value pairs.

    if {![string match [set tmp [$mytree set -node id1 value1]] "value1"]} {
	error "Failed to get expected return from set -node:  $tmp != value1"
    }

    if {![string match [set tmp [$mytree set -node id2 value2]] "value2"]} {
	error "Failed to get expected return from set -node:  $tmp != value2"
    }

    if {![string match [set tmp [$mytree get -node id1]] "value1"]} {
	error "Failed to get expected return from get -node:  $tmp != value1"
    }

    if {![string match [set tmp [$mytree get -node id2]] "value2"]} {
	error "Failed to get expected return from get -node:  $tmp != value2"
    }

    if {![string match [set tmp [$mytree get -best id2]] "value2"]} {
	error "Failed to get expected return from get -best:  $tmp != value2"
    }

    if {![string match [set tmp [$mytree get id2]] "value2"]} {
	error "Failed to get expected return from get:  $tmp != value2"
    }

    set fail [catch {$mytree get -tree id2} tmp]

    if {!$fail || [string match $tmp "value2"]} {
	error "Got un-expected return from get -tree:  $tmp != value2"
    }
    
    # Add some children

    if {![string match [set child [$mytree add]] "tree0.node.2"]} {
	error "Failed to get expected return from add:  $child != tree0.node.2"
    }

    if {![string match [set child1 [$mytree add]] "tree0.node.3"]} {
	error "Failed to get expected return from add:  $child1 != tree0.node.3"
    }

    if {![string match [set child1_1 [$child1 add]] "tree0.node.4"]} {
	error "Failed to get expected return from add:  $child1_1 !=\
	  tree0.node.4"
    }

    if {![string match [set child1_2 [$child1 add]] "tree0.node.5"]} {
	error "Failed to get expected return from add:  $child1_2 !=\
	  tree0.node.5"
    }
    
    # Test the key search
    if {![string match [set tmp [$child1_2 search "id2"]] "$mytree"]} {
	error "Failed to get expected return from search:  $tmp !=\
	  $mytree"
    }

    # Test more set/get with multiple word values and keys.

    if {![string match [set t [$child1 set -node id6 "this's a multiword\
      string"]] "this's a multiword string"]} {
	error "Failed to get expected return from set -node:  $t != this's a\
	  multiword string"
    }

    if {![string match [set t [$child1 get -node id6]] "this's a multiword\
      string"]} {
	error "Failed to get expected return from get -node:  $t != this's a\
	  multiword string"
    }

    if {![string match [set t [$child1 set -node "multi word id"\
      "multiword data"]] "multiword data"]} {
	error "Failed to get expected return from set -node:  $t != multiword\
	  data"
    }

    if {![string match [set t [$child1 get -node "multi word id"]]\
      "multiword data"]} {
	error "Failed to get expected return from get -node:  $t != multiword\
	  data"
    }

    if {![string match [set t [$child1 set -tree "multi word id"\
      "TREE multiword data"]] "TREE multiword data"]} {
	error "Failed to get expected return from set -node:  $t != TREE\
	  multiword data"
    }

    if {![string match [set t [$child1 get -node "multi word id"]]\
      "multiword data"]} {
	error "Failed to get expected return from get -node:  $t != multiword\
	  data"
    }

    if {![string match [set t [$child1 get -best "multi word id"]]\
      "multiword data"]} {
	error "Failed to get expected return from get -best:  $t != multiword\
	  data"
    }

    if {![string match [set t [$child1 get "multi word id"]]\
      "multiword data"]} {
	error "Failed to get expected return from get:  $t != multiword data"
    }

    if {![string match [set t [$child1 get -tree "multi word id"]]\
      "TREE multiword data"]} {
	error "Failed to get expected return from get:  $t != TREE multiword\
	  data"
    }

    if {![string match [set t [$child1_1 get -best "multi word id"]]\
      "TREE multiword data"]} {
	error "Failed to get expected return from get: -best  $t != TREE\
	  multiword data"
    }

    if {![string match [set t [$child1_1 get "multi word id"]] "TREE multiword\
      data"]} {
	error "Failed to get expected return from get:  $t != TREE multiword\
	  data"
    }


    # Test the tree dump 

    set expectDump {{ NODE: tree0.node.1 - primary: tree0 - idNum: 1}\
      {  Parent: None} {  Values:} {    tree0.node.1.id1 value1}\
      {    tree0.node.1.id2 value2} {  Children: tree0.node.2 tree0.node.3}\
      {     NODE: tree0.node.2 - primary: tree0 - idNum: 2} {      Parent:\
      tree0.node.1} {      treeGetNodeValueList: Node tree0.node.2 has no\
      values} {      Children: None} {     NODE: tree0.node.3 - primary: tree0\
      - idNum: 3} {      Parent: tree0.node.1} {      Values:}\
      {        tree0.node.3.multi word id multiword data}\
      {        tree0.node.3.id6 this's a multiword string} {      Children:\
      tree0.node.4 tree0.node.5} {         NODE: tree0.node.4 - primary: tree0\
      - idNum: 4} {          Parent: tree0.node.3}\
      {          treeGetNodeValueList: Node tree0.node.4 has no values}\
      {          Children: None} {         NODE: tree0.node.5 - primary: tree0\
      - idNum: 5} {          Parent: tree0.node.3}\
      {          treeGetNodeValueList: Node tree0.node.5 has no values}\
      {          Children: None}}
    set dump [$mytree dump -r]

    if {$dump != $expectDump} {

	for {set i 0} {$i < [llength $dump]} {incr i} {
	    puts "[string match [lindex $dump $i] [lindex $expectDump $i]]: \n\
	      [lindex $dump $i] \n   [lindex $expectDump $i] "
	}
	error "Failed to get expected return from dump:  ..$dump..\n !=\
	  \n..$expectDump..\n"

    }

    # Test childList

    if {![string match [set t [$child1 childList]]\
      "tree0.node.4 tree0.node.5"]} {
	error "Failed to get expected return from childList:  $t !=\
	  tree0.node.4 tree0.node.5"
    }
    
    # Test parent

    if {![string match [set t [$child1_1 parent]] $child1]} {
	error "Failed to get expected return from parent:  $t != $child1"
    }
    
    # test valueList

    set expectValueList {{{tree0.node.3.multi word id} {multiword data}}\
      {tree0.node.3.id6 {this's a multiword string}}}
    if {![string match [set t [$child1 valueList]] $expectValueList]} {
	error "Failed to get expected return from valueList:  $t !=\
	  $expectValueList"
    }

    # test base

    if {![string match [set t [$child1 base]] $mytree]} {
	error "Failed to get expected return from base:  $t != $mytree"
    }

    # test methods of deleting nodes and trees

    $child1 delete -r

    set fail [catch {puts "[$child1_2 set -node id4]"}]
    if {!$fail} {
	error "Successful set in a deleted node"
    }

    set fail [catch {puts "[$child1 set -node id4]"}]
    if {!$fail} {
	error "Successful set in a child of a deleted node"
    }

    $mytree destroyTree
    set fail [catch {puts "[$child set -tree id3]"}]
    if {!$fail} {
	error "Set an entry in a deleted node!"
    }
    
    # Confirm a new tree can be constructed.

    set fail [catch {newTreeObject} mytree2]
    if {$fail} {
	error "failed to recreate a destroyed tree!"
    }

    if {![string match [set strange [newTreeObject]] "tree2.node.1"]} {
	error "Failed to get expected return from newTreeObject:  $t !=\
	  $tree2.node.1"
    }

    puts "tree Object tests OK - No Errors"

}

if {$argv == "-testObj"} {
    TestTreeObj
}
